;The G.E.C.K. PowerUp for New Vegas - Part of the library
;A simple limited printf function
;File version 0.9
;Written by Fedor Uvarov, 2010

;Supported format identifiers:
;%s - inserts a string. Corresponds to address of a null-terminated string.
;%d or %d - inserts a signed decimal integer. Corresponds to said number.
;%x - inserts an unsigned hexadecimal number using lower case letters. Corresponds to said number.
;%X - the same but with upper-case letters
;%% - inserts a % sign. No corresponding arguments.
;%<digit 1-9><modifier> - inserts something but pads it with spaces to given length. More than 1 digit padding is not supported yet.
;%0<digit 1-9><modifier> - the same but pads with 0's
;%Anything else - the next format argument is skipped and the format modifier is printed literally

macro PRINTF_DATA {
PrintfDigits db "0123456789", 0
PrintfHexDigits db "0123456789abcdef", 0
PrintfHexUpperDigits db "0123456789ABCDEF", 0
}

printf_StartAddrOffset = 24
printf_localvar_1 = 0
printf_localvar_2 = 4

macro PRINTF_PROCS {
;Constructs the string in stack.
;On return: EAX=size of the resulting string, ECX=how much space was used (EAX+all local vars+return address+args)
;It has one argument: stack address, but I'm not making it a proc since I don't want to cripple stack frame
sprintf_instack:
	;How we use memory:
	;StartAddr - address to next arg there
	;esi - current char in format string
	;edi - points to the first byte BEFORE the resulting string
	;esp - current symbol in the result
	;bx - current filling setting. bl - symbol, bh - number of chars
	;bl will be set to non-null by percent sign, so it's also a flag.
	;eax, ecx, edx - used for anything
	push esi
	push edi
	push ebx
	push 0	;An extra memory location
	push 0	;Another one
	mov edi, esp
	
	mov esi, [edi+printf_StartAddrOffset]
	mov esi, [esi]
	xor ebx, ebx
	
	sprintf_instack.process_printf:
	mov al, [esi]
	.if bl	;We're in a format specifier
		.if al = '%'	;Percent sign
			dec esp
			mov [esp], al
			xor bl, bl
		.elseif (al = 'd') | (al = 'i')	;A decimal integer
			add DWORD [edi+printf_StartAddrOffset], 4
			_IntToStr
			xor bl, bl
		.elseif al = '0'
			mov bl, '0'
		.elseif al = 'x'
			add DWORD [edi+printf_StartAddrOffset], 4
			_HexToStr PrintfHexDigits
			xor bl, bl
		.elseif al = 'X'
			add DWORD [edi+printf_StartAddrOffset], 4
			_HexToStr PrintfHexUpperDigits
			xor bl, bl
		.elseif al = 's'
			add DWORD [edi+printf_StartAddrOffset], 4
			_CopyString
			xor bl, bl
		.else	;A digit or an unknown special char
			;Testing if it's a digit
			mov edx, PrintfDigits
			xor ecx, ecx
			sprintf_instack.testifdigit:
			mov cl, [edx]
			cmp cl, al
			je sprintf_instack.yesdigit
			test cl, cl
			jz sprintf_instack.notdigit
			inc edx
			jmp sprintf_instack.testifdigit
			
			sprintf_instack.yesdigit:
			sub edx, PrintfDigits
			mov bh, dl
			jmp sprintf_instack.enditeration
			
			sprintf_instack.notdigit:
			;Not a digit
			dec esp
			mov [esp], BYTE '%'
			.if bh
				.if ~ (bl = ' ')
					dec esp
					mov [esp], bl
				.endif
				mov ecx, PrintfDigits
				xor edx, edx
				mov dl, bh
				add ecx, edx
				dec esp
				mov cl, [ecx]
				mov [esp], cl
			.endif
			dec esp
			mov [esp], al
			xor bx, bx
			add DWORD [edi+printf_StartAddrOffset], 4	;Because it probably had an associated arg
		.endif
	.else	;Not a format specifier yet
		.if al = '%'
			mov bl, ' '
		.else
			dec esp
			mov [esp], al
		.endif
	.endif
	sprintf_instack.enditeration:
	inc esi
	test al, al
	jnz sprintf_instack.process_printf
	
	;Reversing the result
	mov ebx, esp
	mov eax, edi	;Last byte BEFORE result
	sub eax, esp
	xor edx, edx
	mov ecx, 2
	div ecx
	
	.if eax
		mov ecx, eax
		mov edx, edi
		dec edx
		sprintf_instack.reverse_result: mov al, [esp]
		mov ah, [edx]
		mov [edx], al
		mov [esp], ah
		dec edx
		inc esp
		loop sprintf_instack.reverse_result
	.endif
	mov esp, ebx
	
	;Calculating sizes
	mov eax, edi	;Last byte BEFORE result
	sub eax, esp	;Last byte of result
	mov ecx, eax
	add ecx, 28	;5 local DWORDS, return address, 1 argument
	
	;Restoring registers
	mov ebx, [edi+8]
	mov esi, [edi+16]
	push DWORD [edi+20]
	mov edi, [edi+12]
	retn


;Just like sprintf_instack bud aligns the stack to 4 bytes.
;It does NOT pad it with 0's any longer
;EDX will contain pointer to the real result.
;ECX will be set properly
sprintf_instack_align:
	;Saving data and calling sprintf_instack
	push ebx	;+4 bytes in stack
	push esi	;+4 bytes in stack
	mov esi, esp
	push DWORD [esp+12]	;This will be counted by the callee
	call sprintf_instack	;Return address counted by the callee
	mov edx, esp	;Saving real start of the address
	
	;Aligning
	;Counting how much to add
	mov ebx, esp
	and esp, 0FFFFFFFCh
	sub ebx, esp
	
	;Changing registers and setting edx
	add ecx, ebx	;Padding
	add ecx, 16	;2 locals + return addess + arg
	
	;Restoring ESI and EBX
	push DWORD [esi+8]	;Return address
	mov ebx, [esi+4]
	mov esi, [esi]
	retn
	
;Executes sprintf with padding on StartAddr, then calls ProcToCall
;ProcToCall must accept 3 arguments: ArgToPass, address of the resulting string, size of it
;All registers will be left as the called function leaves them except
;EBX (restored as it was) and ESP (restored to the beginning of the stack)
proc printf_call StartAddr, ProcToCall, ArgToPass
	push [ArgToPass]
	push [ProcToCall]
	mov ebx, esp
	stdcall sprintf_instack_align, [StartAddr]
	push ecx	;How much to clean
	push eax
	push edx
	push DWORD [ebx+4]
	call DWORD [ebx]
	
	;Restoring everything
	add esp, ecx
	add esp, 8	;Own "locals"
	ret
endp
}


;Converts a signed decimal integer to a string (%d or %i)
;[edi+printf_StartAddrOffset] must contain the integer
macro _IntToStr {
	mov eax, [edi+printf_StartAddrOffset]
	mov eax, [eax]
	mov [edi+printf_localvar_1], esp
	
	;Checking sign
	local dec_sign_end
	mov [edi+printf_localvar_2], DWORD 0
	test eax, -1
	jns dec_sign_end
	mov [edi+printf_localvar_2], DWORD 1
	imul eax, -1
	.if bh
		dec bh
	.endif
	dec_sign_end:
	
	;Converting
	local dec_cvtloop
	dec_cvtloop:
	mov ecx, 10
	xor edx, edx
	div ecx
	add edx, PrintfDigits
	dec esp
	mov cl, BYTE [edx]
	mov [esp], cl
	.if bh
		dec bh
	.endif
	test eax, eax
	jnz dec_cvtloop
	
	;If we have padding with 0, we first pad then add sign
	;With space (or anything else) it's vice-versa
	.if bl = '0'
		_pad_string
		_add_sign
	.else
		_add_sign
		_pad_string
	.endif
	
	;Preparing to reverse
	mov eax, DWORD [edi+printf_localvar_1]
	sub eax, esp
	xor edx, edx
	mov ecx, 2
	div ecx
	
	;Reversing
	local dec_reverseloop
	.if eax
		mov ecx, eax
		mov edx, DWORD [edi+printf_localvar_1]
		dec edx
		mov [edi+printf_localvar_1], esp
		
		dec_reverseloop:
		mov ah, [edx]
		mov al, [esp]
		mov [edx], al
		mov [esp], ah
		dec edx
		inc esp
		loop dec_reverseloop
		
		;Restoring esp
		mov esp, [edi+printf_localvar_1]
	.endif
	
	mov al, 1	;To continue processing
}

;Converts an unsigned hex integer to string using selected DigitSet (%x or %X)
;(to choose between uppercase and lowercase)
macro _HexToStr DigitSet {
	mov eax, [edi+printf_StartAddrOffset]
	mov eax, [eax]
	mov [edi+printf_localvar_1], esp
	
	local hex_convert
	hex_convert:
	xor ecx, ecx
	mov cl, al
	and cl, 0Fh
	add ecx, DigitSet
	mov cl, [ecx]
	dec esp
	mov [esp], cl
	.if bh
		dec bh
	.endif
	shr eax, 4
	test eax, eax
	jnz hex_convert
	
	_pad_string
	
	;Preparing to reverse
	mov eax, [edi+printf_localvar_1]
	sub eax, esp
	xor edx, edx
	mov ecx, 2
	div ecx
	
	;Reversing
	local hex_reverseloop
	.if eax
		mov ecx, eax
		mov edx, [edi+printf_localvar_1]
		dec edx
		mov [edi+printf_localvar_1], esp
		
		hex_reverseloop:
		mov al, [esp]
		mov ah, [edx]
		mov [esp], ah
		mov [edx], al
		dec edx
		inc esp
		loop hex_reverseloop
		
		mov esp, [edi+printf_localvar_1]
	.endif
	
	mov al, 1	;To continue processing
}

;Copies a string (%s)
macro _CopyString {
	mov ecx, [edi+printf_StartAddrOffset]
	mov ecx, [ecx]
	mov [edi+printf_localvar_1], esp
	
	local copy_loop
	local copy_endofloop
	copy_loop:
	mov al, [ecx]
	test al, al
	jz copy_endofloop
	.if bh
		dec bh
	.endif
	dec esp
	mov [esp], al
	inc ecx
	jmp copy_loop
	copy_endofloop:
	
	;Padding is special
	local copy_padding
	.if bh
		mov ecx, [edi+printf_localvar_1]
		mov [edi+printf_localvar_1], esp
		sub ecx, esp
		
		.if ecx
			xor edx, edx
			mov dl, bh
			xor bh, bh
		
			copy_shifting:
			mov al, [esp]
			sub esp, edx
			mov [esp], al
			add esp, edx
			mov [esp], bl
			inc esp
			loop copy_shifting
		.endif
			
		mov esp, [edi+printf_localvar_1]
	.endif
	
	mov al, 1
}

;Padding
macro _pad_string {
	local dec_padding
	.if bh
		dec_padding:
		dec esp
		mov [esp], bl
		dec bh
		test bh, bh
		jnz dec_padding
	.endif
}

;Adding sign if neccessary
macro _add_sign {
mov ecx, [edi+printf_localvar_2]
.if ecx
	dec esp
	mov [esp], BYTE '-'
.endif
}